#include "obj_loader.cuh"
#include <iostream>
#include <fstream>
#include <sstream>


//#define GLEW_STATIC
//#include <GL/glew.h> 
//#include <GLFW/glfw3.h>


bool loadOBJ(const std::string& filename, Mesh& mesh) {
    std::ifstream file(filename);
    if (!file.is_open()) {
        std::cerr << "Error: Cannot open OBJ file: " << filename << std::endl;
        return false;
    }

    std::string line;
    while (std::getline(file, line)) {
        std::istringstream sstream(line);
        std::string type;
        sstream >> type;

        if (type == "v") {
            Vector vertex;
            sstream >> vertex[0] >> vertex[1] >> vertex[2];
            mesh.vertices.push_back(vertex);
        }
        else if (type == "vt") {
            Vec2 texcoord;
            sstream >> texcoord.u >> texcoord.v;
            mesh.texcoords.push_back(texcoord);
        }
        else if (type == "vn") {
            Vector normal;
            sstream >> normal[0] >> normal[1] >> normal[2];
            mesh.normals.push_back(normal);
        }
        else if (type == "f") {
            Triangle triangle;
            char slash;

            for (int i = 0; i < 3; i++) {
                sstream >> triangle.vertexIndices[i] >> slash >> triangle.texcoordIndices[i] >> slash >> triangle.normalIndices[i];
                triangle.vertexIndices[i]--;
                triangle.texcoordIndices[i]--;
                triangle.normalIndices[i]--;
            }
            mesh.triangles.push_back(triangle);
        }
    }

    return true;
}

void bounding_box(const Mesh& mesh, Precision& x_min, Precision& y_min, Precision& z_min, Precision& x_max, Precision& y_max, Precision& z_max) {

    x_max = y_max = z_max = std::numeric_limits<Precision>::lowest();
    x_min = y_min = z_min = std::numeric_limits<Precision>::max();

    for (int i = 0; i < mesh.vertices.size(); i++) {

        Vector vertex = mesh.vertices[i];

        if (vertex.x() < x_min) x_min = vertex.x();
        if (vertex.y() < y_min) y_min = vertex.y();
        if (vertex.z() < z_min) z_min = vertex.z();

        if (vertex.x() > x_max) x_max = vertex.x();
        if (vertex.y() > y_max) y_max = vertex.y();
        if (vertex.z() > z_max) z_max = vertex.z();
    }
}

//void renderOBJ(Mesh& mesh) {
//    // trojkaty
//    glColor4f(1.0f, 1.0f, 0.0f, 1.0f);  // kolor
//    glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);  // wypelnienie
//    glBegin(GL_TRIANGLES);
//    for (const auto& triangle : mesh.triangles) {
//        for (int i = 0; i < 3; i++) {
//            // normalne
//            if (!mesh.normals.empty()) {
//                const Vector& normal = mesh.normals[triangle.normalIndices[i]];
//                glNormal3f(normal[0], normal[1], normal[2]);
//            }
//
//            // tekstury, na wszelki wypadek, jakbym chcial je uwzglednic w przyszlosci
//            if (!mesh.texcoords.empty()) {
//                const Vec2& texcoord = mesh.texcoords[triangle.texcoordIndices[i]];
//                glTexCoord2f(texcoord.u, texcoord.v);
//            }
//
//            // wierzcholki
//            const Vector& vertex = mesh.vertices[triangle.vertexIndices[i]];
//            glVertex3f(vertex[0], vertex[1], vertex[2]);
//        }
//    }
//    glEnd();
//
//    // krawedzie
//    glColor4f(0.0f, 0.0f, 0.0f, 1.0f);  // kolor
//    glPolygonMode(GL_FRONT_AND_BACK, GL_LINE); 
//    glLineWidth(5.0f);  // grubosc
//    glBegin(GL_TRIANGLES);
//    for (const auto& triangle : mesh.triangles) {
//        for (int i = 0; i < 3; i++) {
//            if (!mesh.normals.empty()) {
//                const Vector& normal = mesh.normals[triangle.normalIndices[i]];
//                glNormal3f(normal[0], normal[1], normal[2]);
//            }
//            const Vector& vertex = mesh.vertices[triangle.vertexIndices[i]];
//            glVertex3f(vertex[0], vertex[1], vertex[2]);
//        }
//    }
//    glEnd();
//
//    glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
//}

// Funkcja pomocnicza do obliczenia iloczynu wektorowego
Vector cross(const Vector& a, const Vector& b) {
    return Vector(
        a[1] * b[2] - a[2] * b[1],
        a[2] * b[0] - a[0] * b[2],
        a[0] * b[1] - a[1] * b[0]
    );
}

// Funkcja pomocnicza do obliczenia iloczynu skalarnego
Precision dot(const Vector& a, const Vector& b) {
    return a[0] * b[0] + a[1] * b[1] + a[2] * b[2];
}

// Funkcja do obliczenia objtoci siatki
Precision calculateVolume(const Mesh& mesh) {
    Precision volume = 0.0f;

    for (const auto& triangle : mesh.triangles) {
        // Pobierz wierzchoki trjkta
        const Vector& v0 = mesh.vertices[triangle.vertexIndices[0]];
        const Vector& v1 = mesh.vertices[triangle.vertexIndices[1]];
        const Vector& v2 = mesh.vertices[triangle.vertexIndices[2]];

        // Oblicz objto czworocianu
        Vector crossProduct = cross(v1 - v0, v2 - v0);
        Precision tetraVolume = dot(v0, crossProduct) / 6.0f;

        // Dodaj objto do sumy
        volume += tetraVolume;
    }

    // Zwr warto absolutn, poniewa objto moe by ujemna w zalenoci od orientacji trjktw
    return std::abs(volume);
}