#include "sdf.cuh"

#include <iostream>
#include <cmath>
#include <algorithm>
#include <array>



// znajdywanie najblizszego punktu na trojkacie do danego wezla
// https://stackoverflow.com/questions/2924795/fastest-way-to-compute-point-to-triangle-distance-in-3d
// Fedor
// rozszerzone o wyznaczanie wektora normalnego dla najblizszego punktu (wektor normalny nie jest tutaj jednostkowy)

__host__ __device__ ClosestPointResult closest_point_triangle(const Mesh& mesh, const size_t triangle_index, const SDF_Node& node) {

    const Triangle& triangle = mesh.triangles[triangle_index];
    const Vector& a = mesh.vertices[triangle.vertexIndices[0]];
    const Vector& b = mesh.vertices[triangle.vertexIndices[1]];
    const Vector& c = mesh.vertices[triangle.vertexIndices[2]];
    const Vector& p = node.x;

    const Vector& na = mesh.normals[triangle.normalIndices[0]];
    const Vector& nb = mesh.normals[triangle.normalIndices[1]];
    const Vector& nc = mesh.normals[triangle.normalIndices[2]];

    const Vector ab = b - a;
    const Vector ac = c - a;
    const Vector ap = p - a;

    const Precision d1 = ab.dot(ap);
    const Precision d2 = ac.dot(ap);
    if (d1 <= 0.f && d2 <= 0.f) {
        // Closest to vertex A
        return { a, na }; //#1
    }

    const Vector bp = p - b;
    const Precision d3 = ab.dot(bp);
    const Precision d4 = ac.dot(bp);
    if (d3 >= 0.f && d4 <= d3) {
        // Closest to vertex B
        return { b, nb }; //#2
    }

    const Vector cp = p - c;
    const Precision d5 = ab.dot(cp);
    const Precision d6 = ac.dot(cp);
    if (d6 >= 0.f && d5 <= d6) {
        // Closest to vertex C
        return { c, nc }; //#3
    }

    const Precision vc = d1 * d4 - d3 * d2;
    if (vc <= 0.f && d1 >= 0.f && d3 <= 0.f) {
        // Closest to edge AB
        const Precision v = d1 / (d1 - d3);
        Vector closestPoint = a + v * ab;
        Vector interpolatedNormal = (1 - v) * na + v * nb;
        return { closestPoint, interpolatedNormal }; //#4
    }

    const Precision vb = d5 * d2 - d1 * d6;
    if (vb <= 0.f && d2 >= 0.f && d6 <= 0.f) {
        // Closest to edge AC
        const Precision v = d2 / (d2 - d6);
        Vector closestPoint = a + v * ac;
        Vector interpolatedNormal = (1 - v) * na + v * nc;
        return { closestPoint, interpolatedNormal }; //#5
    }

    const Precision va = d3 * d6 - d5 * d4;
    if (va <= 0.f && (d4 - d3) >= 0.f && (d5 - d6) >= 0.f) {
        // Closest to edge BC
        const Precision v = (d4 - d3) / ((d4 - d3) + (d5 - d6));
        Vector closestPoint = b + v * (c - b);
        Vector interpolatedNormal = (1 - v) * nb + v * nc;
        return { closestPoint, interpolatedNormal }; //#6
    }

    // Inside triangle (barycentric combination of A, B, C)
    const Precision denom = 1.f / (va + vb + vc);
    const Precision v = vb * denom;
    const Precision w = vc * denom;
    const Vector closestPoint = a + v * ab + w * ac;
    const Vector interpolatedNormal = (1 - v - w) * na + v * nb + w * nc;
    return { closestPoint, interpolatedNormal }; //#0
}


__host__ __device__ const Precision triangle_distance(const Vector& triangle_point, const SDF_Node& node) {
    //return (node.x - triangle_point).norm();

    const Vector dist = node.x - triangle_point;

    return std::sqrt(dist[0] * dist[0] + dist[1] * dist[1] + dist[2] * dist[2]);
}

__host__ __device__ const Vector vector_distance(const Vector& triangle_point, const SDF_Node& node) {

    const Vector dist = triangle_point - node.x;

    return dist;
}
//__device__ Precision signed_distance_interpolation(const Vector& point, SDF_Node* sdf_grid) {
    
    // trzeba bedzie zmienic jak beda wczytywane dowolne modele (albo normalizowac wspolrzedne wczytywanych modeli na, zeby zaczynaly sie od 0
    //int px = static_cast<int>(point.x() / h);
    //int py = static_cast<int>(point.y() / h);
    //int pz = static_cast<int>(point.z() / h);

    //int index[8] = {
    //    grid_index_grid(px, py, pz), // trzeba zrobic osobna funkcje do lapania indeksu dla sdf
    //    grid_index_grid(px + 1, py, pz),
    //    grid_index_grid(px, py + 1, pz),
    //    grid_index_grid(px, py, pz + 1),
    //    grid_index_grid(px + 1, py + 1, pz),
    //    grid_index_grid(px, py + 1, pz + 1),
    //    grid_index_grid(px + 1, py, pz + 1),
    //    grid_index_grid(px + 1, py + 1, pz + 1)
    //};

    //Precision numerator = 0.0f;
    //Precision denominator = 0.0f;

    //for (int i = 0; i < 8; i++) {
    //    Vector diff = point - sfd_grid[index[i]].x;
    //    Precision dist_squared = diff.transpose() * diff;
    //    if (dist_squared < BSPLINE_EPSILON) {  // If the point is exactly at one of the known points
    //        return sfd_grid[index[i]].signed_dist;
    //    }

    //    Precision weight = 1.0f / dist_squared;
    //    numerator += weight * sfd_grid[index[i]].signed_dist;
    //    denominator += weight;
    //}

    //return numerator / denominator;
//    return 0.0;
//
//};


// nie podoba mi sie to, trzeba bedzie poprawic
// dodatkowo trzeba poprawic, bo get_neighbours moze zwrocic liste mniej niz 6 elementow

void wall_distance_neighbour(Node* grid, const size_t grid_index) {

    const Vector& point = grid[grid_index].x;
    int neighbours[6];
    get_neighbors(grid_index, neighbours);
    Vector vec_dist;
    Precision dist;

    vec_dist = grid[neighbours[0]].vector_dist + Vector(h, 0, 0);
    dist = vec_dist.norm();
    if (dist > grid[grid_index].dist) {
        grid[grid_index].dist = dist;
        grid[grid_index].vector_dist = vec_dist;
    }

    vec_dist = grid[neighbours[1]].vector_dist + Vector(-h, 0, 0);
    dist = vec_dist.norm();
    if (dist > grid[grid_index].dist) {
        grid[grid_index].dist = dist;
        grid[grid_index].vector_dist = vec_dist;
    }

    vec_dist = grid[neighbours[2]].vector_dist + Vector(0, h, 0);
    dist = vec_dist.norm();
    if (dist > grid[grid_index].dist) {
        grid[grid_index].dist = dist;
        grid[grid_index].vector_dist = vec_dist;
    }

    vec_dist = grid[neighbours[3]].vector_dist + Vector(0, -h, 0);
    dist = vec_dist.norm();
    if (dist > grid[grid_index].dist) {
        grid[grid_index].dist = dist;
        grid[grid_index].vector_dist = vec_dist;
    }

    vec_dist = grid[neighbours[4]].vector_dist + Vector(0, 0, h);
    dist = vec_dist.norm();
    if (dist > grid[grid_index].dist) {
        grid[grid_index].dist = dist;
        grid[grid_index].vector_dist = vec_dist;
    }

    vec_dist = grid[neighbours[5]].vector_dist + Vector(0, 0, -h);
    dist = vec_dist.norm();
    if (dist > grid[grid_index].dist) {
        grid[grid_index].dist = dist;
        grid[grid_index].vector_dist = vec_dist;
    }
}