#ifndef SDF_H
#define SDF_H

#include "common.cuh"
#include "obj_loader.cuh"
#include "Node.cuh"
#include "Particle.cuh"
#include "Grid.cuh"



struct SDF_Node {

	Vector dist_vec; // w niektorych sytuacjach mozna traktowac jako normalny (sa zwrocone w te sama strone)
	Precision signed_dist;
    Vector x;   // mozna zaoszczedzic pamiec nie przechowywujac pozycji (pozniej przeliczajac ja na podstawie indeksu i rozdzielczosci)
                // mozna tez polaczyc SDF_grid z Grid

    SDF_Node(Precision x, Precision y, Precision z) {
        this->dist_vec = Vector::Zero();
        this->signed_dist = std::numeric_limits<float>::max();
        this->x = Vector(x, y, z);
    }

};

struct ClosestPointResult {
    Vector closestPoint;
    Vector normal;
};



// na przyszlosc, jesli bedziemy chcieli miec inny grid (np. gestszy lub rzadszy) do kolizji
//struct Node_collision {
//
//};
//
//struct Grid_collision {
//
//};

// znalezienie na trojkacie najblizszego punktu do danego punktu (wezla)
// implementacja przez dostosowanie:
// https://stackoverflow.com/questions/2924795/fastest-way-to-compute-point-to-triangle-distance-in-3d
// Fedor
__host__ __device__ ClosestPointResult closest_point_triangle(const Mesh& mesh, const size_t triangle_index, const SDF_Node& node);

// policzenie odleglosci
__host__ __device__ const Precision triangle_distance(const Vector& triangle_point, const SDF_Node& node);

// wektor wodzacy (od wezla do najblizszego punktu trojkata)
__host__ __device__ const Vector vector_distance(const Vector& triangle_point, const SDF_Node& node);

// policzenie odleglosci od sciany, ale na podstawie sasiednich wezlow, a nie trojkata
const Precision wall_distance_neighbour(const Node* grid, const size_t grid_index);


//__device__ Precision signed_distance_interpolation(const Vector& point, SDF_Node* sdf_grid);

//__host__ __device__ const Precision normal_vector_interpolation(const Vector& point, const Mesh& walls);


//void SDF(Mesh& Walls, std::vector<Node> Grid);

// chyba lepiej zrobic osobna strukture grid do sdf, nie musi wtedy miec tej samej rozdzielczosci co siatka (chociaz moze nie miec sensu zeby miala lepsza)


// 1. odleglosc punktu od trojkata
// - trzeba znalezc, przy pomocy triangle_distance

// 2. odleglosc pomiedzy sasiednimi punktami, a odleglosc od sciany
// - minimum z (odleglosc * wektor do sciany + wektor do tego saiada)

// 3. idziemy po trojkatach
// - szukamy min, max z x, y, z, 
// - x / h, przerzucone na inty (do max dodac 1) - uzyskujemy indeks w grid
// - w petli w szescianie min x, max x, ..., max z liczymy odleglosci grid od trojkatow
// - ustalamy znak na podstawie wektora normalnego
// - atomicmin (pewnie z abs), albo cos takiego

// 4. FSM
// 








#endif